@@ -117,12 +117,18 @@ static void read_boxes_inplace(struct kgem *kgem,
 
 static bool download_inplace(struct kgem *kgem, struct kgem_bo *bo)
 {
+	if (unlikely(kgem->wedged))
+		return true;
+
 	if (!kgem_bo_can_map(kgem, bo))
 		return false;
 
 	if (FORCE_INPLACE)
 		return FORCE_INPLACE > 0;
 
+	if (kgem->can_blt_cpu && kgem->max_cpu_size)
+		return false;
+
 	return !__kgem_bo_is_busy(kgem, bo) || bo->tiling == I915_TILING_NONE;
 }
 
@@ -620,7 +626,7 @@ static bool upload_inplace(struct kgem *kgem,
 			   const BoxRec *box,
 			   int n, int bpp)
 {
-	if (kgem->wedged)
+	if (unlikely(kgem->wedged))
 		return true;
 
 	if (!kgem_bo_can_map(kgem, bo) && !upload_inplace__tiled(kgem, bo))
@@ -990,7 +996,7 @@ static bool upload_inplace__xor(struct kgem *kgem,
 				const BoxRec *box,
 				int n, int bpp)
 {
-	if (kgem->wedged)
+	if (unlikely(kgem->wedged))
 		return true;
 
 	if (!kgem_bo_can_map(kgem, bo))
@@ -1382,6 +1388,8 @@ bool sna_replace(struct sna *sna,
 	     pixmap->drawable.bitsPerPixel,
 	     bo->tiling, busy));
 
+	assert(!sna_pixmap(pixmap)->pinned);
+
 	if (!busy && upload_inplace__tiled(kgem, bo)) {
 		BoxRec box;
 
@@ -1484,6 +1492,8 @@ struct kgem_bo *sna_replace__xor(struct sna *sna,
 	     pixmap->drawable.bitsPerPixel,
 	     bo->tiling));
 
+	assert(!sna_pixmap(pixmap)->pinned);
+
 	if (kgem_bo_is_busy(bo)) {
 		struct kgem_bo *new_bo;
 
